;##############################################################################
;
; Copyright (C) 2012-2017, Michele Cappellari
; E-mail: michele.cappellari_at_physics.ox.ac.uk
;
; Updated versions of the software are available from my web page
; http://purl.org/cappellari/software
;
; If you have found this software useful for your research, I would
; appreciate an acknowledgement to the use of the "LTS_LINEFIT program
; described in Cappellari et al. (2013, MNRAS, 432, 1709), which
; combines the Least Trimmed Squares robust technique of Rousseeuw &
; van Driessen (2006) into a least-squares fitting algorithm which
; allows for errors in both variables and intrinsic scatter."
;
; This software is provided as is without any warranty whatsoever.
; Permission to use, for non-commercial purposes is granted.
; Permission to modify for personal or internal use is granted,
; provided this copyright and disclaimer are included unchanged
; at the beginning of the file. All other rights are reserved.
;
;##############################################################################
;+
; NAME:
;       LTS_LINEFIT
; PURPOSE:
;       Best straight-line *robust* fit to data with errors in
;       both coordinates and fitting for the intrinsic scatter.
;       See Sec.3.2 here http://adsabs.harvard.edu/abs/2013MNRAS.432.1709C
;
; EXPLANATION:
;       Linear Least-squares approximation in one-dimension (y = a + b*x),
;       when both x and y data have errors and allowing for intrinsic
;       scatter in the relation.
;
;       Outliers are iteratively clipped using the extremely robust
;       LTS-FAST technique by Rousseeuw & van Driessen (2006)
;       http://dx.doi.org/10.1007/s10618-005-0024-4
;       See also http://books.google.co.uk/books?id=woaH_73s-MwC&pg=PA15
;
; CALLING SEQUENCE:
;        lts_linefit, x, y, sigx, sigy, par, sig_par, chi_sq, $
;            PLOT=plot, RMS=rms, BAD=bad, /BAYES,
;            CLIP=clip, EPSY=epsy, FRAC=frac, GOOD=good
;            OVERPLOT=overplot, PIVOT=pivot, TEXT=text
;
; MODIFICATION HISTORY:
;       V1.0.0: Michele Cappellari, Oxford, 21 March 2011
;       V1.0.1: Included _EXTRA and OVEPLOT, keywords.
;           MC, Oxford, 28 July 2011
;       V1.0.2: Added PIVOT keyword, MC, Oxford, 1 August 2011
;       V1.0.3: Added CLIP keyword. MC, Oxford, 13 March 2013
;       V1.0.4: Fixed program stop affecting earlier IDL versions.
;           Thanks to Xue-Guang Zhang for reporting the problem
;           and the solution. MC, Turku, 10 July 2013
;       V1.0.5: Scale line spacing with character size in text output.
;           MC, Oxford, 19 September 2013
;       V1.0.6: Check that all input vectors have the same size.
;           MC, Baltimore, 8 June 2014
;       V1.0.7: Text plotting changes. MC, Oxford, 26 June 2014
;       V1.0.8: Minor documentation updates. MC, Oxford, 26 October 2015
;       V1.0.9: Increased upper limit of intrinsic scatter accounting for
;           uncertainty of standard deviation with small samples.
;           Michele Cappellari, Oxford, 26 July 2017
;-
;------------------------------------------------------------------------------
function ltsl_line_residuals, ab, X=x, Y=y, SIGX=sigx, SIGY=sigy
compile_opt idl2, hidden

res = (ab[0] + ab[1]*x - y) / sqrt(sigy^2 + (ab[1]*sigx)^2)

return, res
end
;----------------------------------------------------------------------------
pro ltsl_line_fitting, x, y, sigx, sigy, ab, sig_AB, chi2
compile_opt idl2, hidden

fcnargs = {X:x, Y:y, SIGX:sigx, SIGY:sigy}
ab = MPFIT('ltsl_line_residuals', ab, FUNCTARGS=fcnargs, $
        /QUIET, BESTNORM=chi2, PERROR=sig_AB)

end
;----------------------------------------------------------------------------
pro ltsl_fast_algorithm, x, y, sigx, sigy, h, ab, good
compile_opt idl2, hidden

; Robust least trimmed squares regression.
; Pg. 38 of Rousseeuw & van Driessen (2006)
; http://dx.doi.org/10.1007/s10618-005-0024-4
;
n = n_elements(x)
m = 500 ; Number of random starting points
abV = dblarr(m,2)
chi2V = dblarr(m)
for j=0,m-1 do begin ; Draw m random starting points
    w = (sort(randomu(s,n)))[0:1] ; Reshuffle the list without repetition
    ab = linfit(x[w], y[w]) ; Find a line going trough two random points
    for k=0,2 do begin ; Run C-steps up to H_3
        res = ltsl_line_residuals(ab, X=x, Y=y, SIGX=sigx, SIGY=sigy)
        good = (sort(abs(res)))[0:h-1] ; Fit the h points with smallest errors
        ltsl_line_fitting, x[good], y[good], sigx[good], sigy[good], ab, sig_ab, chi_sq
    endfor
    abV[j,*] = ab
    chi2V[j] = chi_sq
endfor

; Perform full C-steps only for the 10 best results
;
w = sort(chi2v)
nbest = 10
chi_sq = 1d30
for j=0,nbest-1 do begin
    ab1 = abv[w[j],*]
    repeat begin ; Run C-steps to convergence
        abOld = ab1
        res = ltsl_line_residuals(ab1, X=x, Y=y, SIGX=sigx, SIGY=sigy)
        good1 = (sort(abs(res)))[0:h-1] ; Fit the h points with smallest errors
        ltsl_line_fitting, x[good1], y[good1], sigx[good1], sigy[good1], ab1, sig_ab1, chi1_sq
    endrep until array_equal(abOld, ab1)
    if chi_sq gt chi1_sq then begin
        ab = ab1  ; Save best solution
        good = good1
        chi_sq = chi1_sq
    endif
endfor

end
;------------------------------------------------------------------------------
function ltsl_find_outliers, sig_int, $
    X=x, Y=y, SIGX=sigx, SIGY=sigy1, AB=ab, SIG_AB=sig_ab, $
    H=h, GOOD=good, BAD=bad, CHI_SQ=chi_sq, OFFS=offs, CLIP=clip
compile_opt idl2, hidden

sigy = sqrt(sigy1^2 + sig_int^2) ; Gaussian intrinsic scatter

if h eq n_elements(x) then begin ; No outliers detection

    ab = linfit(x, y, MEASURE_ERR=sigy) ; quick initial guess
    ltsl_line_fitting, x, y, sigx, sigy, ab, sig_ab, chi_sq
    bad = -1 ; No outliers

endif else begin ; Robust fit and outliers detection

    ; Initial estimate using the maximum breakdown of
    ; the method of 50% but minimum efficiency
    ;
    ltsl_fast_algorithm, x, y, sigx, sigy, h, ab, good

    ; 99% confidence sigma clipping
    ;
    while 1 do begin

        res = ltsl_line_residuals(ab, X=x, Y=y, SIGX=sigx, SIGY=sigy)
        sig = stddev(res[good])
        goodOld = good
        good = where(abs(res) lt clip*sig, h, COMPLEMENT=bad)
        if array_equal(good, goodOld) then break
        ltsl_line_fitting, x[good], y[good], sigx[good], sigy[good], ab, sig_ab, chi_sq

    endwhile

endelse

; To determine 1sigma error on the intrinsic scatter the chi2
; is decreased by 1sigma=sqrt(2(h-2)) while optimizing (a,b)
;
if offs eq 1 then dchi = sqrt(2d*(h-2d)) else dchi = 0d

print, 'sig_int: ', sig_int, ' ', (chi_sq+dchi)/(h-2d) - 1d, FORMAT='(a,f10.4,a,f10.4)'

return, (chi_sq+dchi)/(h-2d) - 1d
end
;------------------------------------------------------------------------------
pro ltsl_single_fit, x, y, sigx, sigy, par, sig_par, chi_sq, h, good, bad, clip
compile_opt idl2, hidden

print, 'bracketing sig_int solution'
sig1 = 0d ; minimum sig_int is zero
f1 = ltsl_find_outliers(sig1, X=x, Y=y, SIGX=sigx, SIGY=sigy, AB=ab, H=h, GOOD=good, OFFS=0, CLIP=clip)
res = ab[0] + ab[1]*x - y ; Total residuals ignoring measurement errors
std = stddev(res[good])
sig2 = std*(1d + 3d/sqrt(2d*n_elements(good)))  ; Observed scatter + 3sigma error
f2 = ltsl_find_outliers(sig2, X=x, Y=y, SIGX=sigx, SIGY=sigy, H=h, OFFS=0, CLIP=clip)

fargs = {X:x, Y:y, SIGX:sigx, SIGY:sigy, H:h, OFFS:0, CLIP:clip} ; requires Chi2/DOF=1
print, 'Computing sig_int'
sig_int = cap_zbrent(sig1, sig2, f1, f2, FUNC_NAME='ltsl_find_outliers', FUNCTARGS=fargs, SUCCESS=ok, /QUIET)
if ok eq 0 then begin
    message, 'No intrinsic scatter or errors overestimated', /INFO
    sig_int = 0d
    sig_int_err = 0d
endif else begin
    print, 'Computing sig_int error'
    sig1 = sig_int
    f1 = sqrt(2d/(h-2d)) ; By construction chi2=h-2 when sig=sig_int
    f2 += sqrt(2d/(h-2d))
    fargs = {X:x, Y:y, SIGX:sigx, SIGY:sigy, H:h, OFFS:1, CLIP:clip} ; chi2 can always decrease
    sigMax_int = cap_zbrent(sig1, sig2, f1, f2, FUNC_NAME='ltsl_find_outliers', FUNCTARGS=fargs, /QUIET)
    if ok eq 0 then print, 'WARNING: Failed to compute uncertainty on scatter'
    sig_int_err = sigMax_int - sig_int
endelse

; Repeat fit at best-fitting location
;
print, 'Repeat at best fitting solution'
tmp = ltsl_find_outliers(sig_int, X=x, Y=y, SIGX=sigx, SIGY=sigy, $
    AB=ab, SIG_AB=sig_ab, CHI_SQ=chi_sq, H=h, GOOD=good, BAD=bad, OFFS=0, CLIP=clip)
par = [ab[*],sig_int]
sig_par = [sig_ab[*],sig_int_err] ; No formal errors in sig_int

end
;------------------------------------------------------------------------------
pro ltsl_display_errors, par, sig_par, epsy, str
compile_opt idl2, hidden

txt = ['intercept: ','slope: ','scatter: ']
prec = intarr(3)
w = where(sig_par ne 0 and par ne 0)
prec[w] = ceil(alog10(abs(par[w]))) - floor(alog10(sig_par[w])) + 1
dg = strtrim(prec, 2)
for j=0,2 do print, txt[j], par[j], ' +/- ', sig_par[j], FORMAT='(a12,g0.'+dg[j]+',a,g0.2)'

if N_PARAMS() eq 4 then begin
    str = ''
    txt = ['a = ', 'b = ', textoidl('\varepsilon_y = ')]
    ns = n_elements(epsy) ? 2 : 1
    for j=0,ns do str += txt[j] + string(par[j],FORMAT='(g0.'+dg[j]+')') + textoidl(' \pm ') + string(sig_par[j],FORMAT='(g0.2)') + '!C'
endif

end
;------------------------------------------------------------------------------
pro lts_linefit, x0, y, sigx, sigy, par, sig_par, chi_sq, $
    BAD=bad, BAYES=bayes, CLIP=clip1, EPSY=epsy, FRAC=frac, GOOD=good, $
    _EXTRA=ex, OVERPLOT=overplot, PIVOT=pivot, PLOT=plot, RMS=rms, TEXT=text
compile_opt idl2
;on_error, 2

if total(n_elements(x0) eq [n_elements(y), n_elements(sigx), n_elements(sigy)]) ne 3 then $
    message, '[X, Y, SIGX, SIGY] must have rthe same size'

if n_elements(pivot) gt 0 then begin
    x = x0-pivot
endif else begin
    x = x0
    pivot = 0d
endelse
if n_elements(clip1) eq 0 then clip = 2.6 else clip = clip1 ; Default clip=2.6*sig-->99% for Gaussian

p = 2 ; two dimensions
n = N_ELEMENTS(x)
if n_elements(frac) eq 0 then h = (n+p+1)/2 else h = round(frac*n) > (n+p+1)/2

ltsl_single_fit, x, y, sigx, sigy, par, sig_par, chi_sq, h, good, bad, clip
rms = stddev(par[0] + par[1]*x[good] - y[good])

; Bayesian estimate using Kelly (2007) routine
;
if keyword_set(bayes) then begin
    linmix_err, x[good], y[good], post, XSIG=sigx[good], YSIG=sigy[good], /SILENT
    par3 = [median(post.alpha), median(post.beta), median(sqrt(post.sigsqr))]
    sig_par3 = [cap_sigma_interval(post.alpha), cap_sigma_interval(post.beta), cap_sigma_interval(sqrt(post.sigsqr))]
    print, '################################# Bayes values and errors'
    ltsl_display_errors, par3, sig_par3
endif

print, '################################# Formal values and errors'
ltsl_display_errors, par, sig_par, epsy, str
print, 'Observed rms scatter: ', rms, FORMAT='(a,g0.3)'
if pivot ne 0 then print, 'y = a + b*(x-pivot) with pivot = ', pivot, FORMAT='(a,g0.4)'
print, '##########################################################'

if keyword_set(plot) then begin
    ploterror, x0, y, sigx, sigy, /NOHAT, /NODATA, _EXTRA=ex, $
        TITLE=textoidl('Best fit, 1\sigma (68%) and 2.6\sigma (99%)')
    x1 = !x.CRANGE ; Extend lines over full plot
    y1 = par[0] + par[1]*(x1-pivot)
    cgplot, x1, y1, /OVERPLOT
    cgplot, x1, y1 + rms, LINE=2, /OVERPLOT, COLOR='red'
    cgplot, x1, y1 - rms, LINE=2, /OVERPLOT, COLOR='red'
    cgplot, x1, y1 + 2.6*rms, LINE=1, /OVERPLOT, COLOR='red'
    cgplot, x1, y1 - 2.6*rms, LINE=1, /OVERPLOT, COLOR='red'
    oploterror, x0[good], y[good], sigx[good], sigy[good], $
        /NOHAT, PSYM=16, COLOR='blue', ERRCOLOR='blue'
    if bad[0] ne -1 then $
        oploterror, x0[bad], y[bad], sigx[bad], sigy[bad], /NOHAT, $
            PSYM=14, COLOR='forest green', ERRCOLOR='forest green'
    cgplot, x0, y, PSYM=1, SYMSIZE=0.5, THICK=(!P.THICK/2 > 1), /OVERPLOT

    if keyword_set(text) then begin
        str += textoidl('\Delta = ') + string(rms,FORMAT='(g0.2)') + '!C'
        if pivot ne 0 then $
            str += textoidl('(x_0 = ') + string(pivot,FORMAT='(g0.4,")")')
        xy = convert_coord([!x.CRANGE[0],!y.CRANGE[1]],/DATA,/TO_NORMAL)
        charsize = !P.CHARSIZE ? !P.CHARSIZE : 1
        dy = !d.y_ch_size/float(!d.y_size) * charsize * 1.1
        xt = xy[0] + !P.TICKLEN*1.1
        yt = xy[1] - !P.TICKLEN*1.1 - dy
        cgtext, xt, yt, str, /NORMAL
    endif
    if n_elements(overplot) eq 1 then void = execute(overplot)
endif

end
;------------------------------------------------------------------------------
